'''
Created on Nov 2, 2017

@author: Charlie McLouth
'''
import os.path
from flask import Flask, current_app, g
from flask_restful import Api
from p4rest.flask.config import P4RestAppConfig
from p4rest.session import P4RestSession
from p4rest.util import osenviron, P4JSONEncoder, APP_NAME, APP_CONFIG_KEY
from p4rest.flaskapi.debug import Debugging, Debugging2
from p4rest.flaskapi.sessions import P4Sessions
from p4rest.flaskapi.commands import P4Commands


def create_app(defaults=None):
    '''documentation

    export FLASK_APP="p4rest.flask.application:create_app()"
    '''
    Flask.config_class = P4RestAppConfig
    Flask.json_encoder = P4JSONEncoder
    app = Flask(APP_NAME)
    app_path = app.root_path
    # add/update passed defaults
    app.config.update(defaults or {})
    # add/update from config file
    if app.config.get("DEBUG", False):
        if osenviron.get(APP_CONFIG_KEY, 
                         app.config.get(APP_CONFIG_KEY, None)) is None:
            app.config[APP_CONFIG_KEY] = os.path.join(app_path, 
                                                      APP_NAME + ".json")
    configfile = app.config.get_config_file(APP_CONFIG_KEY)
    # add/update program defaults
    app.config.set_program_defaults(os.path.dirname(configfile))
    app.config.from_json(configfile, silent=True)
    # validate application configuration
    app.config.validate(APP_CONFIG_KEY)
    # dump config file
    app.config.dump_to_json(configfile, silent=True)
    # register items
    register_resources(app)
    register_cli(app)
    register_callbacks(app)
    return app
    

def register_resources(app):
    '''documentation'''
    api = Api(app)
    api.add_resource(Debugging, "/debug/", "/debug/<command>")
    api.add_resource(Debugging2, "/debug2/<command>")
    api.add_resource(P4Sessions, "/sessions/<sessionid>", "/sessions/", 
                     endpoint="sessions")
#     api.add_resource(AdminSessions, "/admin/sessions/<sessionid>", "/admin/sessions/", 
#                      endpoint="adminsessions")
    api.add_resource(P4Commands, "/commands/<command>", 
                     endpoint="commands")

def register_cli(app):
    '''documentation'''
    pass

def register_callbacks(app):
    '''register any callbacks'''
    @app.before_first_request
    def init_app():
        '''
        thrown before the first request.
        
        Perform any application initialization here
        '''
        current_app.logger.debug("current_app.before_first_request - enter")
        if current_app.config.from_json_if(APP_CONFIG_KEY):
            current_app.config.validate(APP_CONFIG_KEY)
        current_app.logger.debug("current_app.before_first_request - exit")
    @app.before_request
    def init_request():
        '''
        thrown before the request is processed.
        
        Perform any application initialization here
        '''
        current_app.logger.debug("current_app.before_request - enter")
        if current_app.config.from_json_if(APP_CONFIG_KEY):
            current_app.config.validate(APP_CONFIG_KEY)
        current_app.logger.debug("current_app.before_request - exit")
    @app.teardown_appcontext
    def close_app(error):
        '''
        thrown after the request is completed.
        
        free any resources here
        '''
        current_app.logger.debug("current_app.teardown_appcontext - enter")
        session = getattr(g, "current_session", None)
        if session is not None and isinstance(session, P4RestSession):
            session.dispose()
        current_app.logger.debug("current_app.teardown_appcontext - exit")

if __name__ == '__main__':
    app = create_app({"DEBUG": True})
    app.run(debug=True)
